/* dataCell.h

   Copyright (C) 1999 by David Rose.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   To contact the original author of this program, write to:
           David Rose
           224 W. Dryden St Apt. 418
	   Glendale, CA 91202
	   USA

	   drose@lafn.org
   */

#ifndef DATACELL_H
#define DATACELL_H

#include "tableDataProxy.h"
#include "tableInfo.h"

#include <string>
#include <iostream.h>
#include <time.h>

class DataCell {
public:
  enum Type {
    DC_null,
    DC_bool,
    DC_int,
    DC_float,
    DC_string,
    DC_timestamp,
    DC_elapsed_time,
    DC_foreign_key,
  };

  DataCell();
  DataCell(const DataCell &copy);
  DataCell(const DataCell &copy, Type new_type);
  DataCell &operator = (const DataCell &copy);

  DataCell(int value, bool valid = true);
  DataCell(double value, bool valid = true);
  DataCell(const string &value, bool valid = true);

  // We have to make bool and time values a special case, because the
  // compiler has a hard time differentiating these types from int.
  // The derived classes DataCellBool() and DataCellTime(), below,
  // define constructors that accept these.

  DataCell(const TableDataProxy &proxy, bool valid = true);

  // This constructor accepts a string of whatever type, formatted as
  // read from the database (but without quotation marks), and
  // converts it into the indicated type.  It accepts a (const char *)
  // instead of a string, so that NULL can have a special meaning,
  // distinct from "".
  DataCell(const char *db_value, Type type);

  bool operator == (const DataCell &other) const;
  bool operator != (const DataCell &other) const;

  void clear();
  DataCell as_foreign_key(TableID table_id) const;
  void record_db();

  string q_str() const;
  string e_str() const;

  bool is_null() const;
  Type get_type() const;

  // The following functions return the value stored in the DataCell.
  // They must be matched up to the type as returned by get_type(); it
  // is an error to call a get_* function of the wrong type.

  // The parameter is (except for get_foreign()) the value to be
  // returned if the value of the DataCell is NULL.

  bool get_bool(bool null_value = false) const;
  int get_int(int null_value = 0) const;
  double get_float(double null_value = 0.0) const;
  const string &get_string(const string &null_value = string()) const;
  time_t get_time(time_t null_value = 0) const;
  TableDataProxy get_foreign(TableID table_id) const;

  // get_text() returns the value of the DataCell as a string,
  // regardless of its type.
  string get_text(const string &null_value = string()) const;

  void output(ostream &out) const;

protected:
  Type _type;
  union {
    bool _bool_value;
    int _int_value;
    double _float_value;
    time_t _time_value;
  } _v;

  // These can't be members of the union because C++ classes with
  // member functions can't be placed in a union.
  string _string_value;
  TableDataProxy _foreign_value;
};

class DataCellBool : public DataCell {
public:
  DataCellBool(bool value, bool valid = true);
};

class DataCellTimestamp : public DataCell {
public:
  DataCellTimestamp(time_t value, bool valid = true);
};

class DataCellElapsedTime : public DataCell {
public:
  DataCellElapsedTime(double value, bool valid = true);
};

inline ostream &operator << (ostream &out, const DataCell &dc) {
  dc.output(out);
  return out;
}

// These string concatenation operators are handy for building a SQL
// statement with DataCell values.
inline string operator + (const string &str, const DataCell &dc) {
  return str + dc.q_str();
}

inline string operator + (const DataCell &dc, const string &str) {
  return dc.q_str() + str;
}

ostream &operator << (ostream &out, DataCell::Type type);

#endif
