/* sfPlayer.C

   Copyright (C) 1999 by David Rose.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   To contact the original author of this program, write to:
           David Rose
           224 W. Dryden St Apt. 418
	   Glendale, CA 91202
	   USA

	   drose@lafn.org
   */

#include "system.h"

#include "sfPlayer.h"
#include "format.h"

#include <algorithm>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/soundcard.h>
#include <sys/ioctl.h>

SFPlayer::
SFPlayer(int soundfile_fd, const string &soundfile_name, 
	 const string &audio_device, int block_samples) {
  _is_valid = false;
  _audio_fd = -1;
  _audio_device = audio_device;

  _samples = NULL;
  _block_samples = block_samples;

#ifdef HAVE_AFOPENNAMEDFD
  _sound_file = afOpenNamedFD(soundfile_fd, "r", NULL,
			      soundfile_name.c_str());
#else
  // This is a terrible hack, and will only work on POSIX-compliant
  // systems that define the /dev/fd/* convention for opening up a
  // file descriptor.
  string fd = "/dev/fd/" + format_int(soundfile_fd);
  _sound_file = afOpenFile(fd.c_str(), "r", NULL);
#endif


  if (_sound_file == AF_NULL_FILEHANDLE) {
    cerr << "Unable to read " << soundfile_name << "\n";
    return;
  }

  _sample_rate = afGetRate(_sound_file, AF_DEFAULT_TRACK);
  _num_channels = (int)afGetChannels(_sound_file, AF_DEFAULT_TRACK);
  _num_frames = afGetFrameCount(_sound_file, AF_DEFAULT_TRACK);

  afGetSampleFormat(_sound_file, AF_DEFAULT_TRACK,
		    &_sample_format, &_sample_width);

  if (_sample_width <= 8) {
    _sample_bytes = 1;
  } else if (_sample_width <= 16) {
    _sample_bytes = 2;
  } else {
    _sample_bytes = 4;
  }

  _samples = 
    new unsigned char[_block_samples * _sample_bytes * _num_channels];

  _is_valid = true;
  _current_frame = 0;
}

SFPlayer::
~SFPlayer() {
  delete[] _samples;
}

bool SFPlayer::
is_valid() const {
  return _is_valid;
}

double SFPlayer::
get_length() const {
  return (double)_num_frames / _sample_rate;
}
  
bool SFPlayer::
play(double start, double length, bool verbose) {
  int start_frame = (int)(start * _sample_rate);

  // The end frame might grow absurdly large, if the -p parameter in
  // was absurdly large.  Thus, we store it in a double to guard
  // against overflow.
  double end_frame = _num_frames;
  if (length >= 0.0) {
    end_frame = start_frame + length * _sample_rate;
  }

  if (!open_audio_device()) {
    return false;
  }

  if (_current_frame != start_frame) {
    afSeekFrame(_sound_file, AF_DEFAULT_TRACK, start_frame);
    _current_frame = start_frame;
  }    

  while (_current_frame < end_frame) {
    int samples_read =
      afReadFrames(_sound_file, AF_DEFAULT_TRACK, _samples,
		   (int)min((double)_block_samples, end_frame - _current_frame));
    
    if (samples_read < 0) {
      _is_valid = false;
      return false;
    }
    if (samples_read == 0) {
      // End of file.
      return true;
    }
    
    int bytes_written = 
      write(_audio_fd, _samples, 
	    samples_read * _sample_bytes * _num_channels);
    
    if (bytes_written < 0) {
      perror(_audio_device.c_str());
      return false;
    }
    
    _current_frame += (bytes_written / (_sample_bytes * _num_channels));

    if (verbose) {
      cerr << "time " << (double)_current_frame / _sample_rate 
	   << "   \r";
    }
  }

  return true;
}

bool SFPlayer::
open_audio_device() {
  _audio_fd = open(_audio_device.c_str(), O_WRONLY, 0);
  if (_audio_fd < 0) {
    perror(_audio_device.c_str());
    return false;
  }

  int format;

  // We temporarily assume this is the format of the sound file.
  // Bogus assumption!  Need to insert real code here soon.
  format = AFMT_S16_LE;

  if (ioctl(_audio_fd, SNDCTL_DSP_SETFMT, &format) < 0) {
    perror("SNDCTL_DSP_SETFMT");
    close(_audio_fd);
    _audio_fd = -1;
    return false;
  }

  if (format != AFMT_S16_LE) {
    cerr << "Unable to match format\n";
    close(_audio_fd);
    _audio_fd = -1;
    return false;
  }

  int num_channels = _num_channels;
  if (ioctl(_audio_fd, SNDCTL_DSP_CHANNELS, &num_channels) < 0) {
    perror("SNDCTL_DSP_CHANNELS");
    close(_audio_fd);
    _audio_fd = -1;
    return false;
  }

  int sample_rate = (int)_sample_rate;
  if (ioctl(_audio_fd, SNDCTL_DSP_SPEED, &sample_rate) < 0) {
    perror("SNDCTL_DSP_SPEED");
    close(_audio_fd);
    _audio_fd = -1;
    return false;
  }

  return true;
}
